// Global Tempo
360 => global int bpm => global int last_bpm; // shared global oops
60000/bpm => global int tick;
global int globalClockTick;
global int lastClockTick;
global int play;

// CONSTANTS
32 => global int INT_SIZE;

// Shared Globals
global int reset;
global int boardBits32;
global int boardBits64;
global int lockAmount;
global int ruleNumber;
global int bitFlipper;
global float synthScaler; 
global float synthLPF;
global Event editHappened;

// Public Globals
33 => global int BOARD_SIZE;
8 => global int STATES;
global int lyfeBoard[BOARD_SIZE];
global int rule[STATES];

// Sound Globals
global Synth s;
global Pulse p;


//// INIT
fun void Init() 
{
    // Initialize clock
    0 => reset;
    0 => globalClockTick;
    -1 => lastClockTick;
    1 => play;

    // Initialize values
    1.0 => synthScaler;
    20000.0 => synthLPF;

    // Initialize board
    InitAutomata();
    Play();
    //<<< "Reset" >>>;
}
fun void InitAutomata()
{
    // Don't lock any cells
    0 => lockAmount;
    30 => ruleNumber;

    // Initialize board, zero first, then set inital cell
    for (0 => int j; j < BOARD_SIZE; j++)
    {
        0 => lyfeBoard[j];
    }

    // Set initial rule
    SetRuleNumber(ruleNumber);

    // Set initial board state
    resetBoard();
    PackBoard();
}
fun void resetBoard()
{
    for (0 => int j; j < BOARD_SIZE; j++)
    {
        0 => lyfeBoard[j];
    }
    1 => lyfeBoard[0];
    0 => bitFlipper;
    8 => lockAmount;
}
fun void PrintBoard()
{
    // Print board, for debugging purposes
    string boardString;
    "|" +=> boardString;
    for (0 => int j; j < BOARD_SIZE; j++)
    {
        if (lyfeBoard[j] == 1)
        {
            " * " +=> boardString;
        }
        else
        {
            "   " +=> boardString;
        }
    }
    "|" +=> boardString;

    <<< boardString >>>;
}
fun void PackBoard()
{
    // Bit pack the lyfeBoard boolean array into 2 32 bit integers
    // This will be used to communicate with Unity
    0 => int curr32 => int curr64;
    // Convert the first 32 bits
    for (INT_SIZE - 1 => int i; i >= 0; i--)
    {
        1 <<=> curr32;
        lyfeBoard[i] |=> curr32;
    }
    // Convert the 33rd bit, (has extra code supporting up to 64 bits)
    for (BOARD_SIZE - 1 => int i; i >= INT_SIZE; i--)
    {
        1 <<=> curr64;
        lyfeBoard[i] |=> curr64;
    }
    // Pack the bit packed integers to the global variables
    curr32 => boardBits32;
    curr64 => boardBits64;
    //<<< boardBits64 + " - " + boardBits32 >>>;
}
fun int bitMaskByte()
{
    // Create a bitmask from start to end
    // start <= mask < end
    255 => int mask;
    return boardBits32 & mask;
}


//// HELPER FUNCTIONS
fun void SetRuleNumber(int ruleNumber)
{
    // Convert rule number to rule array bits
    for (0 => int i; i < STATES; i++)
    {
        ruleNumber % 2 => rule[i];
        ruleNumber / 2 => ruleNumber;
    }
}

fun int WindowToBool(int l, int m, int r)
{
    // Convert window of 3 ints, left, middle, and right to rule index
    l * 4 + m * 2 + r => int index;
    return rule[index]; // 0 if dead, 1 if alive
}

fun void UpdateLyfeBoard()
{
    // This function will update the lyfeBoard cells based on the rule
    // This will advance the board to the next timestep
    int tempBoard[BOARD_SIZE];
    // Copy lyfeBoard to tempBoard
    for (0 => int j; j < BOARD_SIZE; j++)
    {
        lyfeBoard[j] => tempBoard[j];
    }

    // TERRYNG MACHINE LOOP LOCK
    if (lockAmount != 0) 
    {
        // From 0 to lockAmount, lock the bits inside and loop them, shift register 
        // if bitFlipper is on, flip the last element before copying
        // Copy the last element to the first element
        if (bitFlipper && tempBoard[BOARD_SIZE - 1])
        {
            // Flip the last bit before looping back to the first
            !tempBoard[lockAmount - 1] => lyfeBoard[0];
        }
        else {
            // Normal copy the last element to the first element
            tempBoard[lockAmount - 1] => lyfeBoard[0];
        }
        // Shift register, shift cells one to the right
        for (lockAmount - 1 => int j; j > 0; j--)
        {
            tempBoard[j-1] => lyfeBoard[j];
        }
    }

    // CEULLAR AUTOMATA RULES
    // From lockAmount to BOARD_SIZE - lockAmount, apply the rules
    // 3 bit sliding window representing the 255 possible rules
    // Check if cell should be dead or alive

    // Update the last element (with wrap around)
    WindowToBool(tempBoard[BOARD_SIZE - 2], tempBoard[BOARD_SIZE - 1], tempBoard[lockAmount]) => lyfeBoard[BOARD_SIZE - 1];
    // If there are more thatn 2 elements that need to be updated
    if (lockAmount != 32)
    {
        // Update the first element (with wrap around)
        WindowToBool(tempBoard[BOARD_SIZE - 1], tempBoard[lockAmount], tempBoard[lockAmount + 1]) => lyfeBoard[lockAmount];
        // Iterate through the rest of the lyfeBoard
        for (lockAmount + 1 => int j; j < BOARD_SIZE - 1; j++)
        {
            WindowToBool(tempBoard[j - 1], tempBoard[j], tempBoard[j + 1]) => lyfeBoard[j];
        }
    }

    PackBoard();
}



//// MAIN LOOP
fun void Play() 
{
    // Increment clock per tick when playing
    // Either in pause or play state
    // Loop between pause and play state
    // Break out when reset flag is set
    while (!reset)
    {
        PackBoard();
        // Playing - CORE LOOP
        while (play && !reset)
        {
            // Update Clock Tick
            globalClockTick => lastClockTick;
            globalClockTick + 1 => globalClockTick;

            // Pass Time
            (tick/2)::ms => now;

            // Update LyfeBoard for next tick
            UpdateLyfeBoard();

            (tick/2)::ms => now;

            // Play the synthesizer note
            bitMaskByte() => int byte;
            if (byte > 0)
            {
                byte * synthScaler => float freq;
                spork ~ s.SynthNote(tick, freq);
            }
            // If first cell is alive, send a pulse
            if (lyfeBoard[0])
            {
                spork ~ p.trigger(tick);
            }
        }
        // Paused
        while (!play && !reset)
        {
            tick::ms => now;
        }
    }
}
fun void ListenForEdit()
{
    while( true )
    {
        // wait for event
        editHappened => now;

        // update the gain and rate
        SetRuleNumber(ruleNumber);
        // reset the lyfeboard
        if (reset == 1)
        {
            resetBoard();
            0 => reset;
        }
        // set the synthLPF
        s.lowpass(synthLPF);
        // set BPM
        if (last_bpm != bpm)
        {
            60000/bpm => tick;
            bpm => last_bpm;
            s.reverb((1.0 - (bpm-120)/480.0) * 0.07 + 0.01);
        }
    }
}

//// MAIN
spork ~ ListenForEdit();
bpm::ms => now; // Just to give Unity a second to breathe
Init();

/* If need multiple instances
spork ~ Init();

while (true)
{
    10::second => now;
}
*/
